from socket import error

class TeamspeakConnectionError(error):
    pass

class TeamspeakConnectionLost(TeamspeakConnectionError):
    pass

class TeamspeakConnectionTelnetEOF(TeamspeakConnectionLost):
    pass

class TeamspeakConnectionFailed(TeamspeakConnectionError):
    pass



# Copyright (c) 2012 Adam Coddington #
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from collections import deque
import logging
from telnetlib import Telnet
from time import time, sleep

from message import Command, MessageFactory


class TeamspeakConnection(Telnet):
    def __init__(self, hostname, port, timeout, pipe_in, pipe_out, keep_alive=30, poll_interval=0.125):
        self.pipe_in = pipe_in
        self.pipe_out = pipe_out

        self.logger = logging.getLogger('teamspeak3.TeamspeakConnection')

        self.commands_unresponded = deque()

        self.keep_alive = keep_alive
        self.poll_interval = poll_interval
        Telnet.__init__(self, hostname, port, timeout)

    def write_command(self, command):
        self.logger.info("Sending command %s" % command.__repr__())
        self.commands_unresponded.append(command)
        self.write("%s\n" % command.output)

    def write_keep_alive(self):
        self.logger.debug("Sending keepalive message.")
        self.write("\n")

    def main_loop(self):
        while True:
            incoming = self.receive_message()
            if incoming:
                self.pipe_out.put(incoming)
            else:
                # Only write messages if we have nothing incoming
                if not self.pipe_in.empty():
                    comm = self.pipe_in.get_nowait()
                    if isinstance(comm, Command):
                        self.write_command(comm)
                elif int(time()) % self.keep_alive == 0:
                    self.write_keep_alive()
            sleep(self.poll_interval)

    def receive_message(self):
        try:
            incoming_message = self.read_until('\n', self.timeout).strip()
            if incoming_message.strip():
                self.logger.debug("Incoming string \"%s\"" % incoming_message)
                message = MessageFactory.get_message(incoming_message)
                if message:
                    if message.is_response():
                        message.set_origination(
                                    self.commands_unresponded.popleft()
                                )
                    elif message.is_reset_message():
                        # Command didn't ask for a response
                        if self.commands_unresponded:
                            self.commands_unresponded.popleft()
                    self.logger.info("Received message %s" % message.__repr__())
                    return message
        except ValueError as e:
            pass
        except IndexError as e:
            self.logger.warning(
                    "Unable to create message for \"%s\"; %s" % (
                        incoming_message, 
                        e
                    )
                )
        except Exception as e:
            return e



# Copyright (c) 2012 Adam Coddington
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import re

__all__ = ['Message', 'Command']

class MessageFactory(object):
    @classmethod
    def get_message(cls, incoming_string):
        if incoming_string[0] != incoming_string[0].upper():
            first_item = incoming_string.split(' ')[0]
            if "=" in first_item and "|" in incoming_string:
                return MultipartMessage(incoming_string)
            else:
                return Message(incoming_string)

class MessageBase(object):
    MAPPINGS = {
            '\\\\': '\\',
            '\\/': '/',
            '\\s': ' ',
            '\\p': '|',
            '\\a': '',
            '\\b': '',
            '\\f': '\n',
            '\\n': '\n',
            '\\r': '\n',
            '\\t': '\t',
            '\\v': '',
            }

    def __eq__(self, other):
        if self.__repr__() == other.__repr__():
            return True
        return False

    def __unicode__(self):
        return unicode(self.__str__())

    def __repr__(self):
        return "<%s>" % self.__str__()

    def _clean_incoming_value_multipart(self, value):
        raw_values = value.split('|')

        items = []
        items.append(
                raw_values[0]
            )
        for raw_item in raw_values[1:]:
            items.append(
                    raw_item.split('=')[1]
                )
        return tuple(items)

    def _clean_incoming_value(self, value):
        return re.sub(r'(\\.)', r'%(\1)s', value) % self.MAPPINGS

    def _clean_outgoing_value(self, value):
        value = str(value)
        for fr, to in self.MAPPINGS.items():
            if to:
                value = value.replace(to, fr)
        return value

    @property
    def ultimate_origination(self):
        if self.is_response():
            return self.origination.command
        else:
            return self.command

class Message(MessageBase):
    def __init__(self, command):
        command = command.strip()
        if not command:
            raise ValueError("No command")

        self.raw_command = command

        self.command = self._get_command_from_string(self.raw_command)
        self.args = self._get_arguments_from_string(self.raw_command)

    def is_reset_message(self):
        if self.command == 'error':
            return True
        return False

    def is_response(self):
        if self.command:
            return False
        return True

    def is_response_to(self, command):
        if self.is_response() and self.origination == command:
            return True
        return False

    def set_origination(self, command):
        self.origination = command

    def _get_command_from_string(self, cmd):
        command = cmd.split(' ')[0]
        if command.find('=') > -1:
            command = None
        return command

    def _get_arguments_from_string(self, cmd):
        args = {}
        raw_args = cmd.split(' ')[1 if self.command else 0:]
        for raw_arg in raw_args:
            arg = raw_arg.split('=', 1)
            attribute = arg[0]
            if len(arg) > 1:
                value = arg[1]
            else:
                value = None
            if value:
                if "|" in value:
                    args[attribute] = self._clean_incoming_value_multipart(value)
                else:
                    args[attribute] = self._clean_incoming_value(value)
            else:
                args[attribute] = None
        return args

    def __getitem__(self, key):
        return self.args[key]

    def keys(self):
        return self.args.keys()

    @property
    def output(self):
        arglist = []
        for param, value in self.args.items():
            arglist.append("%s=%s" % (
                    param,
                    self._clean_outgoing_value(value),
                ))
        if self.is_response():
            return "%s %s" % (
                        self.origination.__repr__(),
                        " ".join(arglist),
                    )
        else:
            return "%s %s" % (
                        self.command,
                        " ".join(arglist),
                    )

    def __str__(self):
        if self.is_response():
            return "%s %s" % (
                        self.origination.__repr__(),
                        self.args
                    )
        else:
            return "%s %s" % (
                        self.command,
                        self.args
                    )

class MultipartMessage(MessageBase):
    def __init__(self, command_string):
        self.command = command_string
        self.origination = None

        self.responses = self.parse_command(
                    self.command
                )

    def parse_command(self, string):
        responses = []
        for string_part in string.split('|'):
            responses.append(
                        Message(
                            string_part
                            )
                    )
        return responses

    def set_origination(self, command):
        self.origination = command
        for response in self.responses:
            response.set_origination(command)

    def __getitem__(self, key):
        return self.responses[key]

    def __str__(self):
        string_list = []
        for response in self.responses:
            string_list.append(repr(response))
        return "[%s]" % (
                ", ".join(string_list)
            )

    def is_response(self):
        return True

    def is_reset_message(self):
        return False

class Command(Message):
    def __init__(self, command, **kwargs):
        self.command = command
        self.args = kwargs



# Copyright (c) 2012 Adam Coddington
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import logging
from multiprocessing import Process, Queue
import socket
from time import sleep

from connection import TeamspeakConnection
from message import Command
from exceptions import TeamspeakConnectionLost, TeamspeakConnectionFailed, \
        TeamspeakConnectionTelnetEOF

__all__ = ['Client']


class Client(object):
    def __init__(self, hostname='127.0.0.1', port=25639, timeout=0.25):
        self.pipe_out = Queue()
        self.pipe_in = Queue()
        self.timeout = timeout

        self.logger = logging.getLogger('teamspeak3.Client')

        self.proc = Process(
                target=self.__class__.start_connection,
                args=(hostname, port, timeout, self.pipe_in, self.pipe_out)
            )
        self.proc.start()
        self.logger.debug("Spawned subprocess %s" % self.proc.pid)

        self._verify_initial_connection()

    def _verify_initial_connection(self):
        command = Command(
                'whoami',
            )
        self.send_command(command)
        attempts = 25
        for attempt in range(attempts):
            message = self.get_message()
            if message and message.is_response_to(command):
                return True
            sleep(self.timeout)
        self.close()
        raise TeamspeakConnectionFailed()

    def __enter__(self, *args, **kwargs):
        return Client(*args, **kwargs)

    def __exit__(self, type, value, traceback):
        self.proc.terminate()
        return True

    def close(self):
        self.logger.debug("Terminating subprocess %s" % self.proc.pid)
        self.proc.terminate()

    @classmethod
    def start_connection(cls, hostname, port, timeout, pipe_in, pipe_out):
        # Connect the out to the in, and the in to the out.
        try:
            conn = TeamspeakConnection(
                        hostname,
                        port,
                        timeout,
                        pipe_out,
                        pipe_in
                    )
            conn.main_loop()
        except socket.error:
            pipe_in.put('TERM')

    def get_messages(self):
        messages = []
        while True:
            message = self.get_message()
            if not message:
                return messages
            messages.append(message)

    def get_message(self):
        try:
            if not self.pipe_in.empty():
                msg = self.pipe_in.get_nowait()
                if isinstance(msg, Exception):
                    raise msg
                elif msg == 'TERM':
                    self.close()
                else:
                    return msg
        except EOFError:
            self.close()
            raise TeamspeakConnectionTelnetEOF()
        if not self.proc.is_alive():
            self.logger.debug("Subprocess %s terminated" % self.proc.pid)
            raise TeamspeakConnectionLost()
        return None

    def subscribe(self, type='any'):
        """
        Shortcut method to subscribe to all messages received from the client.
        """
        return self.send_command(
                    Command(
                            'clientnotifyregister',
                            schandlerid=0,
                            event=type
                        )
                )

    def send_command(self, command):
        if not self.proc.is_alive():
            raise TeamspeakConnectionLost()
        self.pipe_out.put(command)
